import sys
import logging
from pathlib import Path
from collections import defaultdict
from typing import Union
import line_profiler

from jpype import JImplements, JOverride, JString, JClass

import stackprinter

from . import custom_exceptions


_JavaNullPointerException = JClass('java.lang.NullPointerException')

logger = logging.getLogger(__name__)


# *** stacktrace configuration ***
# set stackprinter color style. Default is plaintext. Other choices are darkbg,
# darkbg2, darkbg3, lightbg, lightbg2, lightbg3.
_stackprinter_style = 'plaintext'
# prune tracebacks to only show only show stack levels in the user's py5 code.
_prune_tracebacks = True
_module_install_dir = str(Path(__file__).parent)


_EXCEPTION_MSGS = {
    **custom_exceptions.CUSTOM_EXCEPTION_MSGS,
}


def _exception_msg(exc_type_name, exc_msg, py5info):
    try:
        msg = _EXCEPTION_MSGS.get(exc_type_name, exc_msg)
        if isinstance(msg, str):
            return msg
        elif callable(msg):
            return msg(exc_type_name, exc_msg, py5info)
        else:
            logger.error(
                f'unknown exception msg type for {exc_type_name}: {type(msg).__name__}')
            return exc_msg
    except Exception as e:
        logger.error(
            f'error generating exception msg for {exc_type_name}: {e}')
        return exc_msg


def register_exception_msg(exc_type_name: str, msg: Union[str, callable]):
    _EXCEPTION_MSGS[exc_type_name] = msg


def handle_exception(exc_type, exc_value, exc_tb):
    py5info = []
    try:
        if _prune_tracebacks and hasattr(exc_tb, 'tb_next'):
            prev_tb = exc_tb
            trim_tb = None
            tb = exc_tb.tb_next
            while hasattr(tb, 'tb_next') and hasattr(tb, 'tb_frame'):
                f_code = tb.tb_frame.f_code
                if f_code.co_filename.startswith(_module_install_dir):
                    py5info.append((Path(f_code.co_filename[(len(_module_install_dir) + 1):]).parts,
                                    f_code.co_name))
                    if trim_tb is None:
                        trim_tb = prev_tb
                prev_tb = tb
                tb = tb.tb_next
            if trim_tb:
                trim_tb.tb_next = None
    except Exception as e:
        logger.critical(
            f'Exception thrown while examining error traceback: {str(e)}')

    errmsg = stackprinter.format(
        thing=(exc_type, exc_value, exc_tb.tb_next),
        show_vals='line',
        style=_stackprinter_style,
        suppressed_paths=[r"lib/python.*?/site-packages/numpy/",
                          r"lib/python.*?/site-packages/py5/"])

    if _prune_tracebacks:
        errmsg = errmsg.replace(
            str(exc_value),
            _exception_msg(
                exc_type.__name__,
                str(exc_value),
                py5info))

    logger.critical(errmsg)

    sys.last_type, sys.last_value, sys.last_traceback = exc_type, exc_value, exc_tb


@JImplements('py5.core.Py5Methods')
class Py5Methods:

    def __init__(self, sketch):
        self._sketch = sketch
        self._functions = dict()
        self._pre_hooks = defaultdict(dict)
        self._post_hooks = defaultdict(dict)
        self._profiler = line_profiler.LineProfiler()
        self._is_terminated = False

    def set_functions(self, **kwargs):
        self._functions.update(kwargs)

    def profile_functions(self, function_names):
        for fname in function_names:
            func = self._functions[fname]
            self._profiler.add_function(func)
            self._functions[fname] = self._profiler.wrap_function(func)

    def dump_stats(self):
        self._profiler.print_stats()

    def add_pre_hook(self, method_name, hook_name, hook):
        if self._is_terminated and hasattr(hook, 'sketch_terminated'):
            hook.sketch_terminated()
        else:
            self._pre_hooks[method_name][hook_name] = hook

    def add_post_hook(self, method_name, hook_name, hook):
        if self._is_terminated and hasattr(hook, 'sketch_terminated'):
            hook.sketch_terminated()
        else:
            self._post_hooks[method_name][hook_name] = hook

    def add_pre_hooks(self, method_hooks):
        for method_name, hook_name, hook in method_hooks:
            self.add_pre_hook(method_name, hook_name, hook)

    def add_post_hooks(self, method_hooks):
        for method_name, hook_name, hook in method_hooks:
            self.add_post_hook(method_name, hook_name, hook)

    def remove_pre_hook(self, method_name, hook_name):
        if hook_name in self._pre_hooks[method_name]:
            self._pre_hooks[method_name].pop(hook_name)

    def remove_post_hook(self, method_name, hook_name):
        if hook_name in self._post_hooks[method_name]:
            self._post_hooks[method_name].pop(hook_name)

    def terminate_hooks(self):
        for method_name, hooks in self._pre_hooks.items():
            for hook_name, hook in list(hooks.items()):
                if hasattr(hook, 'sketch_terminated'):
                    hook.sketch_terminated()
                self.remove_pre_hook(method_name, hook_name)
        for method_name, hooks in self._post_hooks.items():
            for hook_name, hook in list(hooks.items()):
                if hasattr(hook, 'sketch_terminated'):
                    hook.sketch_terminated()
                self.remove_post_hook(method_name, hook_name)

    @JOverride
    def get_function_list(self):
        return [JString(s) for s in self._functions.keys()]

    @JOverride
    def run_method(self, method_name, params):
        try:
            if method_name in self._functions:
                # first run the pre-hooks, if any
                if method_name in self._pre_hooks:
                    for hook in list(self._pre_hooks[method_name].values()):
                        hook(self._sketch)

                # now run the actual method
                self._functions[method_name](*params)

                # finally, post-hooks
                if method_name in self._post_hooks:
                    for hook in list(self._post_hooks[method_name].values()):
                        hook(self._sketch)
            return True
        except Exception:
            handle_exception(*sys.exc_info())
            self._sketch._terminate_sketch()
            return False

    @JOverride
    def shutdown(self):
        logger.info('shutdown initiated')
        try:
            self._sketch._shutdown()
            self._is_terminated = True
            self.terminate_hooks()
            logger.info('shutdown complete')
        except Exception:
            logger.exception('exception in sketch shutdown sequence')



import functools

import numpy as np
from jpype import _jcustomizer, JClass


_PVector = JClass('processing.core.PVector')
_PMatrix2D = JClass('processing.core.PMatrix2D')
_PMatrix3D = JClass('processing.core.PMatrix3D')


def _numpy_to_pvector(array):
    if array.shape in [(2,), (3,)]:
        return _PVector(*array.tolist())
    else:
        raise RuntimeError(
            'numpy array is the wrong size to convert to a pvector')


def _numpy_to_pmatrix2d(array):
    return _PMatrix2D(*array.flatten().tolist())


def _numpy_to_pmatrix3d(array):
    return _PMatrix3D(*array.flatten().tolist())


def _numpy_to_pmatrix(jcls, array):
    if array.shape == (2, 3):
        return _numpy_to_pmatrix2d(array)
    elif array.shape == (4, 4):
        return _numpy_to_pmatrix3d(array)
    else:
        raise RuntimeError(
            'numpy array is the wrong size to convert to a pmatrix')


_jcustomizer.JConversion(
    'processing.core.PVector',
    np.ndarray)(_numpy_to_pvector)
_jcustomizer.JConversion(
    'processing.core.PMatrix',
    np.ndarray)(_numpy_to_pmatrix)


def _pvector_to_numpy(pvector):
    return np.array([pvector.x, pvector.y, pvector.z])


def _pmatrix2d_to_numpy(pmatrix2d):
    return np.array([[pmatrix2d.m00, pmatrix2d.m01, pmatrix2d.m02],
                     [pmatrix2d.m10, pmatrix2d.m11, pmatrix2d.m12]])


def _pmatrix3d_to_numpy(pmatrix3d):
    return np.array([[pmatrix3d.m00, pmatrix3d.m01, pmatrix3d.m02, pmatrix3d.m03],
                     [pmatrix3d.m10, pmatrix3d.m11, pmatrix3d.m12, pmatrix3d.m13],
                     [pmatrix3d.m20, pmatrix3d.m21, pmatrix3d.m22, pmatrix3d.m23],
                     [pmatrix3d.m30, pmatrix3d.m31, pmatrix3d.m32, pmatrix3d.m33]])


def _pmatrix_to_numpy(pmatrix):
    if isinstance(pmatrix, _PMatrix2D):
        return _pmatrix2d_to_numpy(pmatrix)
    elif isinstance(pmatrix, _PMatrix3D):
        return _pmatrix3d_to_numpy(pmatrix)
    else:
        raise RuntimeError(
            f'do not know how to convert object of type {type(pmatrix).__name__}')


def _get_matrix_wrapper(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        ret = f(self_)
        if not args:
            return _numpy_to_pmatrix(ret)
        if len(args) == 1:
            target = args[0]
            if (isinstance(target, np.ndarray) and target.shape == (2, 3)
                    and isinstance(ret, _PMatrix2D)):
                target[0, 0] = ret.m00
                target[0, 1] = ret.m01
                target[0, 2] = ret.m02
                target[1, 0] = ret.m10
                target[1, 1] = ret.m11
                target[1, 2] = ret.m12
                return target
            elif (isinstance(target, np.ndarray) and target.shape == (4, 4)
                    and isinstance(ret, _PMatrix3D)):
                target[0, 0] = ret.m00
                target[0, 1] = ret.m01
                target[0, 2] = ret.m02
                target[0, 3] = ret.m03
                target[1, 0] = ret.m10
                target[1, 1] = ret.m11
                target[1, 2] = ret.m12
                target[1, 3] = ret.m13
                target[2, 0] = ret.m20
                target[2, 1] = ret.m21
                target[2, 2] = ret.m22
                target[2, 3] = ret.m23
                target[3, 0] = ret.m30
                target[3, 1] = ret.m31
                target[3, 2] = ret.m32
                target[3, 3] = ret.m33
                return target
            raise RuntimeError(
                "target must be a numpy array that matches the size of processing's matrix")
        raise RuntimeError(
            'unexpected arguments passed to set_matrix function')
    return decorated


def _get_pvector_wrapper(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        ret = f(self_, args[0])
        if len(args) == 1:
            return _pvector_to_numpy(ret)
        if len(args) == 2:
            target = args[1]
            if (isinstance(target, np.ndarray) and target.shape in [(2,), (3,)]
                    and isinstance(ret, _PVector)):
                target[0] = ret.x
                target[1] = ret.y
                if target.shape == (3,):
                    target[2] = ret.z
                return target
            raise RuntimeError(
                "target must be a numpy array that matches the size of processing's pvector")
        raise RuntimeError('unexpected arguments passed to function')
    return decorated



from pathlib import Path

import jpype


_options = []
_classpath = []


def is_jvm_running():
    return jpype.isJVMStarted()


def check_jvm_running():
    if jpype.isJVMStarted():
        raise RuntimeError("the jvm is already running")


def set_options(*options):
    check_jvm_running()
    global _options
    _options = list(options)


def add_options(*options):
    check_jvm_running()
    _options.extend(options)


def get_classpath():
    return jpype.getClassPath()


def add_classpath(classpath):
    check_jvm_running()
    if not isinstance(classpath, Path):
        classpath = Path(classpath)
    jpype.addClassPath(classpath.absolute())


def add_jars(path):
    check_jvm_running()
    if not isinstance(path, Path):
        path = Path(path)
    if path.exists():
        for jarfile in path.glob("**/*.[Jj][Aa][Rr]"):
            jpype.addClassPath(jarfile.absolute())


def start_jvm():
    for c in _classpath:
        print(f'adding {c}')
        jpype.addClassPath(c)
    jpype.startJVM(*_options, convertStrings=False)


__all__ = ['is_jvm_running', 'check_jvm_running',
           'set_options', 'add_options',
           'get_classpath', 'add_classpath',
           'add_jars', 'start_jvm']



from pathlib import Path
from setuptools import setup

with open('README.rst') as f:
    README = f.read()


with open(Path('py5', '__init__.py')) as f:
    for line in f.readlines():
        if line.startswith('__version__'):
            break
    VERSION = line.split("'")[-2]


INSTALL_REQUIRES = [
    'jpype1>=1.0.1',
    'line_profiler>=2.1.2',
    'noise>=1.2.2',
    'nptyping>=1.3',
    'numpy>=1.18',
    'pandas>=1.0',
    'requests>=2.24',
    'stackprinter>=0.2.4',
]

setup(
    name='py5',
    version=VERSION,
    packages=['py5', 'py5.mixins', 'py5_tools', 'py5_tools.tools'],
    py_modules=['setup'],
    package_data={
        "py5": ['jars/*.jar', 'jars/*/*.jar', '*.pyi', 'py.typed']
    },
    python_requires='>3.8',
    install_requires=INSTALL_REQUIRES,
    description='Processing for CPython',
    long_description=README,
    author='Jim Schmitz',
    author_email='jim@ixora.io',
    entry_points={
        'console_scripts': [
            'run_sketch = py5_tools.tools.run_sketch:main',
            'py5cmd = py5_tools.tools.py5cmd:main',
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Artistic Software',
        'Topic :: Multimedia :: Graphics',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Java',
    ],
)



from jpype import _jcustomizer

from .sketch import Sketch, Py5Graphics, Py5Image, Py5Font, Py5Shape, Py5Shader


def init_jpype_converters():
    data = [
        ("processing.core.PImage", Py5Image),
        ("processing.core.PImage", Py5Graphics),
        ("processing.core.PFont", Py5Font),
        ("processing.core.PShape", Py5Shape),
        ("processing.opengl.PShader", Py5Shader),
        ("processing.core.PApplet", Sketch),
    ]

    def convert(jcls, obj):
        return obj._instance

    for javaname, cls_ in data:
        _jcustomizer.JConversion(javaname, cls_)(convert)



# -*- coding: utf-8 -*-
# *** FORMAT PARAMS ***


UPDATE_DYNAMIC_VARIABLES = [
    'args',
    'display_height',
    'display_width',
    'finished',
    'focused',
    'frame_count',
    'height',
    'is_dead',
    'is_dead_from_error',
    'is_ready',
    'is_running',
    'java_platform',
    'java_version',
    'java_version_name',
    'key',
    'key_code',
    'mouse_button',
    'mouse_x',
    'mouse_y',
    'np_pixels',
    'pixel_height',
    'pixel_width',
    'pixels',
    'pmouse_x',
    'pmouse_y',
    'width']
